/*--------------------------------------------------------------------------
 *  Copyright 2008 utgenome.org
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *--------------------------------------------------------------------------*/
//--------------------------------------
// utgb-shell Project
//
// Import.java
// Since: Jan 20, 2009
//
// $URL: http://svn.utgenome.org/utgb/tags/utgb/utgb-shell/utgb-shell-1.2.9/src/main/java/org/utgenome/shell/Import.java $ 
// $Author: leo $
//--------------------------------------
package org.utgenome.shell;

import java.io.File;
import java.io.FileReader;
import java.io.InputStreamReader;
import java.sql.SQLException;

import org.utgenome.format.fasta.FASTA2Db;
import org.utgenome.format.silk.read.ReadDBBuilder;
import org.utgenome.shell.db.bed.BEDDatabaseGenerator;
import org.xerial.util.log.Logger;
import org.xerial.util.opt.Argument;
import org.xerial.util.opt.Option;
import org.xerial.util.opt.OptionParser;

/**
 * import command
 * 
 * @author leo
 * 
 */
// @Usage(command = "> utgb import", description = "import command")
public class Import extends UTGBShellCommandBase {

	private static Logger _logger = Logger.getLogger(Import.class);
	private OptionParser parser = null;

	private enum FileType {
		AUTO, READ, BED, FASTA
	}

	@Option(symbol = "t", longName = "type", description = "specify the input file type: (AUTO, FASTA, READ, BED)")
	private FileType fileType = FileType.AUTO;

	@Argument(index = 0, required = false)
	private String inputFilePath = null;

	@Option(symbol = "o", longName = "output", varName = "DB FILE NAME", description = "output SQLite DB file name")
	private String outputFileName;

	@Option(symbol = "w", longName = "overwrite", description = "overwrite existing DB files")
	private boolean overwriteDB = false;

	private OptionParser getOptionParser() {
		if (parser == null)
			parser = new OptionParser(this);

		return parser;
	}

	@Override
	public void execute(String[] args) throws UTGBShellException {

		File input = null;
		try {
			getOptionParser().parse(args);

			if (inputFilePath == null)
				_logger.info("use STDIN for the input");
			else {
				_logger.info("input file: " + inputFilePath);
				input = new File(inputFilePath);
				if (!input.exists())
					throw new UTGBShellException("file not found: " + inputFilePath);
			}

			_logger.info("file type: " + fileType);

			if (outputFileName == null) {
				// new File("db").mkdirs();

				outputFileName = String.format("%s.sqlite", inputFilePath);
				int count = 1;
				if (!overwriteDB) {
					while (new File(outputFileName).exists()) {
						outputFileName = String.format("%s.sqlite.%d", inputFilePath, count++);
					}
				}
			}

			_logger.info("output file: " + outputFileName);

			switch (fileType) {
			case READ: {
				ReadDBBuilder builder = new ReadDBBuilder(outputFileName);
				if (input != null)
					builder.build(input.toURL());
				else
					builder.build(new InputStreamReader(System.in));
				break;
			}
			case BED: {
				if (input != null)
					BEDDatabaseGenerator.toSQLiteDB(new FileReader(input), outputFileName);
				else
					BEDDatabaseGenerator.toSQLiteDB(new InputStreamReader(System.in), outputFileName);
				break;
			}
			case FASTA:
				if (input != null)
					FASTA2Db.main(new String[] { inputFilePath, "-o", outputFileName });
				else
					FASTA2Db.main(new String[] { "-o", outputFileName });
				break;
			case AUTO:
			default: {
				_logger.warn("file type (-t) must be given");
				break;
			}
			}

		}
		catch (SQLException e) {
			e.printStackTrace();
		}
		catch (Exception e) {
			_logger.error(e);
		}

	}

	@Override
	public String name() {
		return "import";
	}

	public String getDetailedDescription() {
		return loadUsage("help-import.txt");
	}

	public String getOneLinerDescription() {
		return "import a file and create a new database";
	}

	public String getOptionList() {
		return getOptionParser().getUsage();
	}
}
